#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "access/heapam.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/regexp.h"

PG_MODULE_MAGIC;

Datum regexp_matches(PG_FUNCTION_ARGS);
Datum regexp_split(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(regexp_matches);
Datum
regexp_matches(PG_FUNCTION_ARGS)
{
	text *s = PG_GETARG_TEXT_P(0);
	text *p = PG_GETARG_TEXT_P(1);
	bool return_pre_and_post = (PG_NARGS() > 2 ? PG_GETARG_BOOL(2) : false);
	regex_t *cpat;

	cpat = RE_compile_and_cache(p, regex_flavor);
	regmatch_t *pmatch = palloc0(sizeof(regmatch_t) * (cpat->re_nsub + 1));
	if (RE_execute(cpat,
				VARDATA(s),
				VARSIZE(s) - VARHDRSZ,
				cpat->re_nsub + 1,
				pmatch))
	{
		ArrayType * resultarray;
		TupleDesc rettupdesc;
		Oid rettype;
		Datum result;
		Datum elems[cpat->re_nsub];
		bool nulls[cpat->re_nsub];
		/* get text type oid, too lazy to do it some other way */
		Oid param_type = get_fn_expr_argtype(fcinfo->flinfo, 0);
		size_t i;
		int ndims = 1;
		int dims[1] = {cpat->re_nsub};
		int lbs[1] = {1};
		int16 typlen;
		bool typbyval;
		char typalign;
		TypeFuncClass typefunc;

		get_typlenbyvalalign(param_type, &typlen, &typbyval, &typalign);
		typefunc = get_call_result_type(fcinfo, &rettype, &rettupdesc);

		for (i = 0; i < cpat->re_nsub; ++i)
		{
			int so = pmatch[i+1].rm_so,
				eo = pmatch[i+1].rm_eo;

			if (so < 0 || eo < 0)
			{
				elems[i] = 0;
				nulls[i] = true;
			}
			else
			{
				elems[i] = DirectFunctionCall3(text_substr,
						PointerGetDatum(s),
						Int32GetDatum(so + 1),
						Int32GetDatum(eo - so));
				nulls[i] = false;
			}
		}

		resultarray = construct_md_array(elems, nulls, ndims, dims, lbs,
				param_type, typlen, typbyval, typalign);
		if (typefunc == TYPEFUNC_COMPOSITE)
		{
			Datum vals[4];
			bool  valnulls[4] = {true, true, true, true};

			rettupdesc = BlessTupleDesc(rettupdesc);
			if (return_pre_and_post)
			{
				valnulls[0] = valnulls[3] = false;

				vals[0] = DirectFunctionCall3(text_substr,
						PointerGetDatum(s),
						Int32GetDatum(1),
						Int32GetDatum(pmatch[0].rm_so));

				vals[3] = DirectFunctionCall2(text_substr_no_len,
						PointerGetDatum(s),
						Int32GetDatum(pmatch[0].rm_eo + 1));
			}

			vals[1] = DirectFunctionCall3(text_substr,
					PointerGetDatum(s),
					Int32GetDatum(pmatch[0].rm_so + 1),
					Int32GetDatum(pmatch[0].rm_eo - pmatch[0].rm_so));
			valnulls[1] = false;


			vals[2] = PointerGetDatum(resultarray);
			valnulls[2] = false;

			result = HeapTupleGetDatum(
					heap_form_tuple (rettupdesc, vals, valnulls));
		}
		else
		{
			result = PointerGetDatum(resultarray);
		}

		pfree(pmatch);
		PG_RETURN_DATUM(result);
	}
	/* if no match, fall through and return null */

	PG_RETURN_NULL();
}


typedef struct regexp_split_ctx {
	text		* orig_str;
	size_t		  orig_len;
	pg_wchar	* wide_str;
	size_t		  wide_len;
	regex_t		* cpattern;
	regmatch_t	  match;
	size_t		  offset;
} regexp_split_ctx;

PG_FUNCTION_INFO_V1(regexp_split);
Datum
regexp_split(PG_FUNCTION_ARGS)
{
	FuncCallContext  *funcctx;
	MemoryContext     oldcontext;
	regexp_split_ctx *splitctx;

	if (SRF_IS_FIRSTCALL())
	{
		funcctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		splitctx = (regexp_split_ctx *) palloc(sizeof(regexp_split_ctx));

		splitctx->orig_str = PG_GETARG_TEXT_P_COPY(0);
		splitctx->orig_len = VARSIZE(splitctx->orig_str) - VARHDRSZ;

		/* TODO: other compile options? */
		splitctx->cpattern = RE_compile_and_cache(PG_GETARG_TEXT_P(1), regex_flavor);

		splitctx->wide_str = (pg_wchar *) palloc((splitctx->orig_len + 1) * sizeof(pg_wchar));
		splitctx->wide_len = pg_mb2wchar_with_len(VARDATA(splitctx->orig_str), splitctx->wide_str, splitctx->orig_len);

		splitctx->offset = 0;

		splitctx->match.rm_so = -1;
		/* both < 0 but not equal */
		splitctx->match.rm_eo = -2;

		MemoryContextSwitchTo(oldcontext);
		funcctx->user_fctx = (void*)splitctx;

		/* Avoid run-away function by making sure we never iterate more than the length of the text */
		funcctx->max_calls = 3 * splitctx->orig_len;
	}

	funcctx = SRF_PERCALL_SETUP();
	splitctx = (regexp_split_ctx*)(funcctx->user_fctx);

	if (funcctx->call_cntr > funcctx->max_calls)
	{
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("set returning split function terminated after iterating %d times", funcctx->call_cntr)));
		SRF_RETURN_DONE(funcctx);
	}

	if (splitctx->offset < splitctx->orig_len)
	{
		regmatch_t * pmatch = &(splitctx->match);
		bool first_match = (pmatch->rm_so < 0 || pmatch->rm_eo < 0);
		bool incremented_offset = false;
		do
		{
			if (pmatch->rm_so == pmatch->rm_eo)
			{
				splitctx->offset++;
				incremented_offset = true;
			}

			if (RE_wchar_execute(splitctx->cpattern,
						splitctx->wide_str,
						splitctx->wide_len,
						splitctx->offset,
						1,
						pmatch))
			{
				if (!((first_match && pmatch->rm_so == pmatch->rm_eo) || (pmatch->rm_so == splitctx->offset && !incremented_offset)))
				{
					Datum result = DirectFunctionCall3(text_substr,
							PointerGetDatum(splitctx->orig_str),
							Int32GetDatum(splitctx->offset + (incremented_offset ? 0 : 1)),
							Int32GetDatum(pmatch->rm_so - splitctx->offset + (incremented_offset ? 1 : 0)));
					splitctx->offset = pmatch->rm_eo;
					SRF_RETURN_NEXT(funcctx, result);
				}
			}
			else
			{
				/* no more matches, return rest of string */
				Datum result = DirectFunctionCall2(text_substr_no_len,
						PointerGetDatum(splitctx->orig_str),
						Int32GetDatum(splitctx->offset + (incremented_offset ? 0 : 1)));
				/* so we know we're done next time thru */
				splitctx->offset = splitctx->orig_len;
				SRF_RETURN_NEXT(funcctx, result);
			}
			first_match = false;
		} while (1);
	}

	SRF_RETURN_DONE(funcctx);
}

